# INTERFACE TO GNUPG PACKAGE
#
# Package for reading and writing text files that may or may not be encrypted.
# File will be encrypted if file path ends in a GPG extension.

# License {{{1
# Copyright (C) 2016-18 Kenneth S. Kundert
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see http://www.gnu.org/licenses/.


# Imports {{{1
from .config import get_setting, override_setting, setting_path
from .error import PasswordError
from .shlib import to_path, cp
from .utilities import OSErrors
from inform import (
    conjoin, cull, display, log, narrate, os_error, warn, is_str,
    full_stop
)
import gnupg
try:
    from StringIO import StringIO   # python2
except ImportError:
    from io import StringIO         # python3

# Globals {{{1
GPG_EXTENSIONS = ['.gpg', '.asc']
ARMOR_CHOICES = ['always', 'never', 'extension']
ActivePythonFile= None

# get_active_python_file() {{{1
def get_active_python_file():
    return ActivePythonFile

# GnuPG class {{{1
class GnuPG(object):
    def __init__(self, path):
        self.path = to_path(path)

    @classmethod
    def initialize(cls,
        gpg_path=None, gpg_home=None, armor=None
    ):

        cls.gpg_path = to_path(
            gpg_path if gpg_path else get_setting('gpg_executable')
        )
        override_setting('gpg_executable', cls.gpg_path)

        cls.gpg_home = to_path(
            gpg_home if gpg_home else get_setting('gpg_home')
        )
        override_setting('gpg_home', cls.gpg_home)

        armor = armor if armor is not None else get_setting('gpg_armor')
        if armor not in ARMOR_CHOICES:
            warn(
                "'%s' is not valid, choose from %s." % (
                    armor, conjoin(ARMOR_CHOICES)
                ), culprit=setting_path('gpg_armor')
            )
            armor = 'extension'
        cls.armor = armor
        override_setting('gpg_armor', armor)

        gpg_args = {}
        if cls.gpg_path:
            gpg_args.update({'gpgbinary': str(cls.gpg_path)})
        if cls.gpg_home:
            gpg_args.update({'gnupghome': str(cls.gpg_home)})
        cls.gpg = gnupg.GPG(**gpg_args)

    def save(self, contents, gpg_ids=None):
        path = self.path
        if not gpg_ids:
            gpg_ids = get_setting('gpg_ids', [])
        if is_str(gpg_ids):
            gpg_ids = gpg_ids.split()
        if not gpg_ids:
            # raise PasswordError('must specify GPG ID.')
            log('no gpg id is available, using symmetric encryption.')

        use_gpg, use_armor = self._choices()
        if use_gpg:
            try:
                encoded = contents.encode(get_setting('encoding'))
                if gpg_ids:
                    encrypted = self.gpg.encrypt(encoded, gpg_ids, armor=use_armor)
                else:
                    encrypted = self.gpg.encrypt(encoded, None, symmetric='AES256', armor=use_armor)
                if not encrypted.ok:
                    msg = ' '.join(cull([
                        'unable to encrypt.',
                        getattr(encrypted, 'stderr', None)
                    ]))
                    raise PasswordError(msg, culprit=path, sep='\n')
                else:
                    path.write_bytes(encrypted.data)
            except ValueError as e:
                raise PasswordError(full_stop(e), culprit=path)
        else:
            path.write_text(contents, encoding=get_setting('encoding'))
        self.chmod()

    def read(self):
        path = self.path
        # file is only assumed to be encrypted if path has gpg extension
        if path.suffix.lower() in GPG_EXTENSIONS:
            try:
                with path.open('rb') as f:
                    decrypted = self.gpg.decrypt_file(f)
                    if not decrypted.ok:
                        msg = ' '.join(cull([
                            'unable to decrypt.',
                            getattr(decrypted, 'stderr', None)
                        ]))
                        raise PasswordError(msg, culprit=path, sep='\n')
            except ValueError as e:
                raise PasswordError(full_stop(e), culprit=path)
            except OSErrors as e:
                raise PasswordError(os_error(e))
            return decrypted.data.decode(get_setting('encoding'))
        else:
            return path.read_text(encoding=get_setting('encoding'))

    def _choices(self):
        extension = self.path.suffix.lower()
        if extension in GPG_EXTENSIONS:
            if self.armor == 'never':
                return True, False
            elif self.armor != 'always' and extension == '.gpg':
                return True, False
            return True, True
        else:
            return False, False

    def remove(self):
        self.path.unlink()

    def backup(self, extension):
        """Creates a backup copy of the file.

        The name of the new file has the specified extension prepended to the
        existing suffixes.
        """
        if extension:
            # prepend extension to list of suffixes
            suffixes = self.path.suffixes
            stem = self.path.stem.partition('.')[0]  # remove all suffixes
            new = to_path(self.path.parent, ''.join([stem, extension] + suffixes))
            self.backup_path = new

            cp(self.path, new)
            return new

    def restore(self):
        "Restores the backup copy of the file."
        cp(self.backup_path, self.path)

    def will_encrypt(self):
        return self.path.suffix in GPG_EXTENSIONS

    def exists(self):
        return self.path.exists()

    def chmod(self, mode=0o600):
        self.path.chmod(mode)

    def __str__(self):
        return str(self.path)


# BufferedFile class {{{1
class BufferedFile(GnuPG):
    def __init__(self, path, ignore_errors_on_close=False):
        # file will only be encrypted if path has gpg extension
        self.path = path
        self.stream = StringIO()
        self.ignore_errors_on_close = ignore_errors_on_close

    def write(self, content):
        self.stream.write(content)

    def flush(self):
        pass

    def close(self):
        if get_setting('discard_logfile'):
            # this is the case when the output would be uninteresting (such as
            # with help messages) and running GPG (and thus risk requiring the
            # user to type in the passphrase) is silly.
            return

        contents = self.stream.getvalue()
        try:
            self.save(contents, get_setting('gpg_ids'))
        except PasswordError:
            if not self.ignore_errors_on_close:
                raise


# PythonFile class {{{1
class PythonFile(GnuPG):
    def run(self):
        global ActivePythonFile
        ActivePythonFile = self.path
        path = self.path
        self.encrypted = path.suffix in ['.gpg', '.asc']
        narrate('reading.', culprit=path)
        try:
            self.code = self.read()
                # need to save the code for the new command
        except OSErrors as e:
            raise PasswordError(os_error(e))

        try:
            compiled = compile(self.code, str(path), 'exec')
        except SyntaxError as e:
            culprit = (e.filename, e.lineno)
            if e.text is None or e.offset is None:
                raise PasswordError(full_stop(e.msg), culprit=culprit)
            else:
                raise PasswordError(
                    e.msg + ':', e.text.rstrip(), (e.offset-1)*' ' + '^',
                    culprit=culprit, sep='\n'
                )
                # File "/home/ken/.config/avendesora/config", line 18
                #   'c': 'google-chrome %s'
                #      ^

        contents = {}
        try:
            exec(compiled, contents)
        except Exception as e:
            from .utilities import error_source
            raise PasswordError(full_stop(e), culprit=error_source())
        ActivePythonFile = None
        return contents

    def create(self, contents, gpg_ids=None):
        path = self.path
        try:
            # check to see if file already exists
            if path.exists():
                # file creation (init) requested, but file already exists
                # don't overwrite the file, instead read it so the information 
                # can be used to create any remaining files.
                display("%s: already exists." % path)
                return

            # create the file
            display('%s: creating.' % path)
            if path.suffix in ['.gpg', '.asc']:
                narrate('encrypting.', culprit=path)
                # encrypt it
                if not gpg_ids:
                    raise PasswordError('gpg_ids missing.')
                self.save(contents, gpg_ids)
            else:
                narrate('not encrypting.', culprit=path)
                # file is not encrypted
                with path.open('wb') as f:
                    f.write(contents.encode(get_setting('encoding')))
        except OSErrors as e:
            raise PasswordError(os_error(e))



from .account import Account, StealthAccount
from .charsets import (
    exclude, LOWERCASE, UPPERCASE, LETTERS, DIGITS, ALPHANUMERIC,
    HEXDIGITS, PUNCTUATION, SYMBOLS, WHITESPACE, PRINTABLE, DISTINGUISHABLE
)
from .error import PasswordError
from .generator import PasswordGenerator
from .obscure import Hide, Hidden, GPG
try:
    from .obscure import Scrypt
except ImportError:  # pragma: no cover
    pass
try:
    from .otp import OTP
except ImportError:  # pragma: no cover
    pass
from .recognize import (
    RecognizeAll, RecognizeAny, RecognizeTitle, RecognizeURL, RecognizeCWD,
    RecognizeHost, RecognizeUser, RecognizeEnvVar, RecognizeNetwork,
    RecognizeFile
)
from .script import Script
from .secrets import (
    Password, Passphrase, PIN, Question, MixedPassword, PasswordRecipe,
    BirthDate, SecretExhausted
)

# the following are used when generating the documentation and are not needed
# otherwise
from . import command
from .account import AccountValue

__version__ = '1.12.0'
__released__ = '2019-01-17'



# Usage {{{1
"""
Avendesora Collaborative Password Generator

Generates passwords and pass phrases based on stored account information.

Usage:
    avendesora [options] <command> [<args>...]

Options:
    -h, --help        output basic usage information.

Commands:
{commands}

Use 'avendesora help <command>' for information on a specific command.
Use 'avendesora help' for list of available help topics.

Documentation can be found at avendesora.readthedocs.io.
"""

# License {{{1
# Copyright (C) 2016-18 Kenneth S. Kundert
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses/.


# Imports {{{1
from .command import Command
from .config import read_config, get_setting
from .error import PasswordError
from .gpg import GnuPG, BufferedFile
from .utilities import OSErrors
from . import shlib
from inform import Inform, Error, done, fatal, output, terminate, os_error
from docopt import docopt
import sys


# Main {{{1
def main():
    try:
        # read config file
        read_config()

        # read command line
        cmdline = docopt(
            __doc__.format(commands=Command.summarize()),
            options_first=True
        )

        # start logging
        logfile = BufferedFile(get_setting('log_file'), True)
        Inform(
            logfile=logfile, hanging_indent=False,
            stream_policy='header', notify_if_no_tty=True
        )
        shlib.set_prefs(use_inform=True, log_cmd=True)

        # run the requested command
        Command.execute(cmdline['<command>'], cmdline['<args>'])
        done()
    except KeyboardInterrupt:
        output('\nTerminated by user.')
        terminate()
    except (PasswordError, Error) as e:
        e.terminate()
    except OSErrors as e:
        fatal(os_error(e))
    done()



# Avendesora Password Generator Settings
#
# Access user settings.

# License {{{1
# Copyright (C) 2016-18 Kenneth S. Kundert
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses/.

# Imports {{{1
from .error import PasswordError
from .preferences import CONFIG_DEFAULTS, NONCONFIG_SETTINGS
from inform import comment, warn, is_str
from pathlib import Path

# Globals {{{1
Config = {}


# read_config() {{{1
def read_config():
    if Config.get('READ'):
        return  # already read

    # First open the config file
    from .gpg import PythonFile
    path = get_setting('config_file')
    assert path.suffix.lower() not in ['.gpg', '.asc']
    config_file = PythonFile(path)
    if not config_file.exists():
        # have not yet initialized this account
        return
    try:
        contents = config_file.run()
        for k, v in contents.items():
            if k.startswith('_'):
                continue
            if k not in CONFIG_DEFAULTS:
                warn('%s: unknown.' % k, culprit=config_file)
                continue
            if k.endswith('_executable'):
                argv = v.split() if is_str(v) else list(v)
                path = Path(argv[0])
                if not path.is_absolute():
                    warn(
                        'should use absolute path for executables.',
                        culprit=(config_file, k)
                    )
            Config[k] = v
        Config['READ'] = True
    except PasswordError:
        comment('not found.', culprit=config_file)

    # Now open the hashes file
    hashes_file = PythonFile(get_setting('hashes_file'))
    try:
        contents = hashes_file.run()
        Config.update({k.lower(): v for k,v in contents.items()})
    except PasswordError:
        pass

    # Now open the account list file
    account_list_file = PythonFile(get_setting('account_list_file'))
    try:
        contents = account_list_file.run()
        Config.update({k.lower(): v for k,v in contents.items()})
    except PasswordError:
        pass

    # initilize GPG
    from .gpg import GnuPG
    GnuPG.initialize()

    # Now read the user key file
    user_key_file = get_setting('user_key_file')
    if user_key_file:
        user_key_file = PythonFile(get_setting('user_key_file'))
        try:
            contents = user_key_file.run()
            Config.update({k.lower(): v for k,v in contents.items()})
        except PasswordError:
            pass


# add_setting() {{{1
def add_setting(name, default):
    CONFIG_DEFAULTS[name] = default


# get_setting() {{{1
def get_setting(name, default=None, expand=True):
    name = name.lower()
    try:
        value = Config[name]
    except KeyError:
        try:
            value = CONFIG_DEFAULTS[name]
        except KeyError:
            try:
                value = NONCONFIG_SETTINGS[name]
            except KeyError:
                return default
    if value is None:
        return default
    if name == 'gpg_ids':
        value = value.split() if is_str(value) else value
    elif name == 'settings_dir':
        value = Path(value)
    elif expand and name.endswith('_file'):
        value = get_setting('settings_dir') / value
    return value


# setting_path() {{{1
def setting_path(name, index=None):
    # this returns a path to a particular setting that is useful as a culprit
    if index is None:
        return (get_setting('config_file'), name)
    else:
        return (get_setting('config_file'), name, index)


# override_setting() {{{1
def override_setting(name, value):
    Config[name.lower()] = value



# pathlib -- extends system pathlib

# License {{{1
# Copyright (C) 2016-2019 Kenneth S. Kundert
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see [http://www.gnu.org/licenses/].

__version__ = '0.4.0'

# Imports {{{1
from pathlib import Path, PosixPath
import codecs
import six
import os
import sys


# is_readable {{{1
def _is_readable(path):
    """
    Tests whether path exists and is readable.

    >>> from extended_pathlib import Path
    >>> Path('/usr/bin/python').is_readable()
    True

    """
    return os.access(str(path), os.R_OK)

PosixPath.is_readable = _is_readable

# is_writable {{{1
def _is_writable(path):
    """
    Tests whether path exists and is writable.

    >>> Path('/usr/bin/python').is_writable()
    False

    """
    return os.access(str(path), os.W_OK)

PosixPath.is_writable = _is_writable

# is_executable {{{1
def _is_executable(path):
    """
    Tests whether path exists and is executable.

    >>> Path('/usr/bin/python').is_executable()
    True

    """
    return os.access(str(path), os.X_OK)

PosixPath.is_executable = _is_executable

# is_hidden {{{1
def _is_hidden(path):
    """
    Tests whether path exists and is hidden.

    >>> Path('/usr/bin/python').is_hidden()
    False

    """
    return path.exists() and str(path).startswith('.')

PosixPath.is_hidden = _is_hidden

# path_from {{{1
def _path_from(path, start):
    """
    Returns relative path from start as a path object.
    This differs from Path.relative_to() in that relative_to() will not return a
    path that starts with '..'.

    >>> Path('.').path_from('..')
    PosixPath('tests')

    """
    return Path(os.path.relpath(str(path), str(start)))

PosixPath.path_from = _path_from

# sans_ext {{{1
def _sans_ext(path):
    """
    Removes the file extension.
    This differs from Path.stem, which returns the final path component
    stripped of its extension. This returns the full path stripped of its
    extension.

    >>> Path('a/b.c').sans_ext()
    PosixPath('a/b')

    """
    return path.parent / path.stem

PosixPath.sans_ext = _sans_ext

# Python 3.5 extensions {{{1
if sys.version_info < (3, 5):

    # read_bytes {{{2
    def _read_bytes(self):
        """
        Open the file in binary mode, read it, and close the file.
        """
        with self.open(mode='rb') as f:
            return f.read()

    PosixPath.read_bytes = _read_bytes

    # read_text {{{2
    def _read_text(self, encoding=None, errors=None):
        """
        Open the file in text mode, read it, and close the file.
        """
        with self.open(mode='r', encoding=encoding, errors=errors) as f:
            return f.read()

    PosixPath.read_text = _read_text

    # write_bytes {{{2
    def _write_bytes(self, data):
        """
        Open the file in binary mode, write it, and close the file.
        """
        if not isinstance(data, six.binary_type):
            raise TypeError(
                'data must be %s, not %s.' % (
                    six.binary_type.__name__, data.__class__.__name__
                )
            )
        with self.open(mode='wb') as f:
            return f.write(data)

    PosixPath.write_bytes = _write_bytes

    # write_text {{{2
    def _write_text(self, text, encoding=None, errors=None):
        """
        Open the file in text mode, write it, and close the file.
        """
        path = str(self)
        with codecs.open(path, mode='w', encoding=encoding, errors=errors) as f:
            return f.write(text)

    PosixPath.write_text = _write_text

    # expand_user {{{2
    def _expanduser(self):
        """
        Return a new path with expanded ~ and ~user constructs.
        """

        path = str(self)
        if path[:1] == '~':
            path = os.path.expanduser(path)
            self = Path(path)
        return self

    PosixPath.expanduser = _expanduser




# Editors
#
# Open and editor on a file.

# License {{{1
# Copyright (C) 2016-18 Kenneth S. Kundert
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses/.

# Imports {{{1
from .config import get_setting, add_setting, setting_path
from .error import PasswordError
from .shlib import Run
from inform import log, os_error

# Editor base class {{{1
class Editor(object):
    pass

# GenericEditor class {{{1
class GenericEditor(Editor):
    NAME = 'generic'

    @classmethod
    def run(cls, filepath, account = None):
        args = dict(
            filepath = filepath,
            account = account,
        )
        try:
            editor_setting = 'edit_account' if account else 'edit_template'
            cmd = get_setting(editor_setting)
            cmd = [e.format(**args) for e in cmd]

            Run(cmd, 'soeW')
        except KeyError as e:
            raise PasswordError(
                *e.args,
                template='invalid field: {0}',
                culprit=setting_path(editor_setting)
            )



#!/usr/bin/env python

from setuptools import setup

with open('README.rst') as file:
    readme = file.read()

setup(
    name = 'avendesora',
    version = '1.12.0',
    author = 'Ken Kundert and Kale Kundert',
    author_email = 'avendesora@nurdletech.com',
    description = 'A password generator and account manager.',
    long_description = readme,
    url = 'https://avendesora.readthedocs.io',
    download_url = 'https://github.com/kenkundert/avendesora/tarball/master',
    license = 'GPLv3+',
    packages = 'avendesora'.split(),
    entry_points = {
        'console_scripts': ['avendesora = avendesora.main:main'],
    },
    install_requires = [
        'appdirs',
        'arrow',
        'docopt',
        'inform>=1.14',
        #'pygobject',
            # pygobject seems broken in pypi. You can try uncommenting the above
            # and see if it works. Otherwise, do the following:
            #   > git clone git://git.gnome.org/pygobject
            #   > pip3.5 install ./pygobject
            # This install will fail if you do not have the right packages
            # installed in linux. For example, gnome-common is one that you will
            # need. Read the error messages carefully to determine
            # what other packages are needed. Then, be sure to install the
            # development versions. For example, on Fedora23 the libffi, 
            # gobject-introspection and pycairo packages were needed. I
            # installed them with:
            #     dnf install libffi-devel
            #     dnf install gobject-introspection-devel
            #     dnf install python3-cairo-devel
        'python-gnupg>=0.4.3',
            # Be careful.  There's a package called 'gnupg' that's an 
            # incompatible fork of 'python-gnupg'.  If both are installed, the 
            # user will probably have compatibility issues.
        'pyotp',
            # pyotp is optional, it provides one-time-password (OTP) secrets.
        #'scrypt',
            # scrypt is optional. If you install it then Avendesora will offer
            # it. It is not required because it is little used and installing it
            # involves compiling C code and so significant additional
            # dependencies such as gcc.
    ],
    setup_requires = 'pytest-runner>=2.0'.split(),
    tests_require = 'pytest pexpect'.split(),
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*',
    keywords = 'avendesora password XKCD'.split(),
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Security :: Cryptography',
        'Topic :: Utilities',
    ],
)



# Commands

# License {{{1
# Copyright (C) 2016-18 Kenneth S. Kundert
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses/.


# Imports {{{1
from .collection import Collection
from .config import get_setting, override_setting
from .editors import GenericEditor
from .error import PasswordError
from .generator import PasswordGenerator
from .gpg import GnuPG, PythonFile
from .obscure import ObscuredSecret
from .shlib import chmod, cp, rm, to_path
from .utilities import query_user, two_columns, OSErrors
from .writer import get_writer
from inform import (
    codicil, columns, cull, display, error, full_stop, output, warn,
    conjoin, join, os_error, is_collection, indent, render,
)
from docopt import docopt
from textwrap import dedent
import re
import sys


# Utilities {{{1
def title(text):
    return full_stop(text[0].upper() + text[1:])


# Command base class {{{1
class Command(object):
    @classmethod
    def commands(cls):
        for cmd in cls.__subclasses__():
            assert is_collection(cmd.NAMES)
            yield cmd
            # currently only one level of subclassing is supported

    @classmethod
    def commands_sorted(cls):
        for cmd in sorted(cls.commands(), key=lambda c: c.get_name()):
            yield cmd

    @classmethod
    def find(cls, name):
        for command in cls.commands():
            if name in command.NAMES:
                return command

    @classmethod
    def execute(cls, name, args):
        if args is None:
            args = []
        command = cls.find(name)
        if not command:
            # no recognizable command was specified
            # in this case, run 'credentials' if one argument is given
            # and 'value' otherwise
            args = [name] + args
            name = 'value' if len(args) > 1 else 'credentials'
            command = cls.find(name)
            if not command:
                error('unknown command.', culprit=name)
                codicil("Use 'avendesora help' for list of available commands."),
                return
        command.run(name, args)

    @classmethod
    def summarize(cls, width=16):
        summaries = []
        for cmd in Command.commands_sorted():
            summaries.append(two_columns(', '.join(cmd.NAMES), cmd.DESCRIPTION))
        return '\n'.join(summaries)

    @classmethod
    def get_name(cls):
        return cls.NAMES[0]

    @classmethod
    def help(cls):
        text = dedent("""
            {title}

            {usage}
        """).strip()
        return text.format(
            title=title(cls.DESCRIPTION), usage=cls.USAGE,
        )

    @classmethod
    def get_help_url(cls):
        try:
            anchor = '-'.join(e.lower() for e in [cls.NAMES[0]] + cls.DESCRIPTION.split())
            return '/commands.html#' + anchor
        except (AttributeError, TypeError):
            pass


# Add {{{1
class Add(Command):
    NAMES = 'add', 'a'
    DESCRIPTION = 'add a new account'
    USAGE = dedent("""
        Usage:
            avendesora add [options] [<template>]
            avendesora a   [options] [<template>]

        Options:
            -f <file>, --file <file>
                                    Add account to specified accounts file.

        Creates a new account starting from a template. The template consists of
        boilerplate code and fields. The fields take the from _NAME_. They
        should be replaced by appropriate values or deleted if not needed. If
        you are using the Vim editor, it is preconfigured to jump to the next
        field when you press 'n'.  If the field is surrounded by '<<' and '>>',
        as in '<<_ACCOUNT_NUMBER_>>', the value you enter will be concealed.

        You can create your own templates by adding them to 'account_templates'
        in the ~/.config/avendesora/config file.

        You can change the editor used when adding account by changing the
        'edit_template', also found in the ~/.config/avendesora/config file.
    """).strip()

    @classmethod
    def help(cls):
        text = dedent("""
            {title}

            {usage}

            The default template is {default}. The available templates are:
            {templates}

            The available accounts files are (the default is given first):
            {files}
        """).strip()

        def indented_list(l):
            indent = get_setting('indent')
            return indent + ('\n'+indent).join(sorted(l))

        return text.format(
            title=title(cls.DESCRIPTION), usage=cls.USAGE,
            default=get_setting('default_account_template'),
            templates=indented_list(get_setting('account_templates').keys()),
            files=indented_list(get_setting('accounts_files', [])),
        )

    @classmethod
    def run(cls, command, args):
        # read command line
        cmdline = docopt(cls.USAGE, argv=[command] + args)

        # determine the accounts file
        prefix = cmdline['--file']
        if prefix:
            candidates = [
                p
                for p in get_setting('accounts_files')
                if p.startswith(prefix)
            ]
            if not candidates:
                raise PasswordError('not found.', cuplrit=cmdline['--file'])
            if len(candidates) > 1:
                raise PasswordError(
                    'ambiguous, matches %s.' % conjoin(candidates),
                    cuplrit=prefix
                )
            filename = candidates[0]
        else:
            filename = get_setting('accounts_files')[0]

        try:
            # get the specified template
            templates = get_setting('account_templates')
            if cmdline['<template>']:
                template_name = cmdline['<template>']
            else:
                template_name = get_setting('default_account_template')
            template = dedent(templates[template_name]).strip() + '\n'
        except KeyError:
            raise PasswordError(
                'unknown account template, choose from %s.' % conjoin(
                    sorted(templates.keys())
                ), culprit=template_name
            )

        try:
            # get original contents of accounts file
            path = to_path(get_setting('settings_dir'), filename)
            orig_accounts_file = PythonFile(path)
            orig_accounts = orig_accounts_file.run()
            gpg_ids = orig_accounts.get('gpg_ids')
            orig_accounts_code = orig_accounts_file.code.strip('\n')

            # backup the original file
            orig_accounts_file.backup('.saved')

            # save the template into temp file
            from tempfile import mktemp
            tmpfilename = mktemp(suffix='_avendesora.gpg')
            tmpfile = GnuPG(tmpfilename)
            tmpfile.save(template, get_setting('gpg_ids'))
        except OSErrors as e:
            raise PasswordError(os_error(e))

        # open template in the editor
        try:
            while (True):
                GenericEditor.run(tmpfile.path)

                # read the tmp file and determine if it has changed
                new = tmpfile.read()
                if not new.strip() or new == template:
                    return output('Unchanged, and so ignored.')
                new = tmpfile.read()

                # remove instructions
                new = '\n'.join([
                    l
                    for l in new.split('\n')
                    if not l.startswith('# Avendesora:')
                ]).strip('\n') + '\n'

                # hide the values that should be hidden
                def hide(match):
                    return 'Hidden(%r)' % ObscuredSecret.hide(match.group(1))
                new = re.sub("<<(.*?)>>", hide, new)

                # add new account to the contents
                accounts = orig_accounts_code + '\n\n\n' + new

                # save the new version
                new_accounts_file = PythonFile(path)
                new_accounts_file.save(accounts, gpg_ids)

                # check the changes to see if there are any issues
                try:
                    new_accounts_file.run()
                    break
                except PasswordError as e:
                    error(e)
                    response = query_user('Try again?')
                    if response.lower() not in ['y', 'yes']:
                        raise PasswordError(
                            'Giving up, restoring original file.'
                        )

        except (PasswordError) + OSErrors as e:
            orig_accounts_file.restore()
            if isinstance(e, OSErrors):
                e = os_error(e)
            error(e)
            codicil(
                '',
                'What you entered can be found in %s.' % tmpfilename,
                'Delete it when done with it.', sep='\n'
            )
            return

        # delete the temp file
        try:
            tmpfile.remove()
        except:
            pass


# Archive {{{1
class Archive(Command):
    NAMES = 'archive', 'A'
    DESCRIPTION = 'generates archive of all account information'
    USAGE = dedent("""
        Usage:
            avendesora archive
            avendesora A
    """).strip()

    @classmethod
    def help(cls):
        text = dedent("""
            {title}

            {usage}

            This command creates an encrypted archive that contains all the
            information in your accounts files, including the fully generated
            secrets.  You should never need this file, but its presence protects
            you in case you lose access to Avendesora. To access your secrets
            without Avendesora, simply decrypt the archive file with GPG.  The
            actual secrets will be hidden, but it easy to retrieve them even
            without Avendesora. When hidden, the secrets are encoded in base64.
            You can decode it by running 'base64 -d -' and pasting the encoded
            secret into the terminal.

            When you run this command it overwrites the existing archive. If you
            have accidentally deleted an account or changed a secret, then
            replacing the archive could cause the last copy of the original
            information to be lost. To prevent this from occurring it is a good
            practice to run the 'changed' command before regenerating the
            archive.  It describes all of the changes that have occurred since
            the last time the archive was generated. You should only regenerate
            the archive once you have convinced yourself all of the changes are
            as expected.
        """).strip()
        return text.format(
            title=title(cls.DESCRIPTION), usage=cls.USAGE,
        )

    @classmethod
    def run(cls, command, args):
        # read command line
        cmdline = docopt(cls.USAGE, argv=[command] + args)
        archive_file = get_setting('archive_file')

        # first, save existing archive if it exists
        try:
            previous_archive = get_setting('previous_archive_file')
            if previous_archive and archive_file.is_file():
                rm(previous_archive)
                cp(archive_file, previous_archive)
        except OSErrors as e:
            raise PasswordError(os_error(e))

        # run the generator
        generator = PasswordGenerator(warnings=False)

        # get dictionary that fully describes the contents of each account
        entries = []
        seen = set()
        for account in generator.all_accounts():
            name = account.get_name()
            if name in seen:
                warn('duplicate account name.', culprit=name)
            seen.add(name)
            entry = account.archive()
            if entry:
                entries.append(indent('%r: %s,' % (name, render(entry)), '    '))

        # build file contents
        from .preferences import ARCHIVE_FILE_CONTENTS
        import arrow
        contents = ARCHIVE_FILE_CONTENTS.format(
            encoding = get_setting('encoding'),
            date = str(arrow.now()),
            accounts = '\n\n'.join(entries)
        )

        archive = GnuPG(archive_file)
        if not archive.will_encrypt():
            warn('archive file is not encrypted.', culprit=archive_file)
        try:
            archive.save(contents)
            chmod(0o600, archive_file)
        except OSErrors as e:
            raise PasswordError(os_error(e), culprit=archive_file)


# Browse {{{1
class Browse(Command):
    NAMES = 'browse', 'b'
    DESCRIPTION = 'open account URL in web browser'
    USAGE = dedent("""
        Usage:
            avendesora browse [options] <account> [<key>]
            avendesora b      [options] <account> [<key>]

        Options:
            -b <browser>, --browser <browser>
                                    Open account in specified browser.
            -l, --list              List available URLs rather than open first.
    """).strip()

    @classmethod
    def help(cls):
        default_browser = get_setting('default_browser')
        browsers = get_setting('browsers')
        browsers = '\n'.join([
            two_columns(k, browsers[k], width=5)
            for k in sorted(browsers)
        ])
        text = dedent("""
            {title}

            {usage}

            The account is examined for URLs, a URL is chosen, and then that URL
            is opened in the chosen browser.  First URLs are gathered from the
            'urls' account attribute, which can be a string containing one or
            more URLs, a list, or a dictionary.  If 'urls' is a dictionary, the
            desired URL can be chosen by entering the key as an argument to the
            browse command. If a key is not given, then the 'default_url'
            account attribute is used to specify the key to use by default. If
            'urls' is not a dictionary, then the first URL specified is used.
            URLs are also taken from RecognizeURL objects in the 'discovery'
            account attribute.  If the 'name' argument is specified, the
            corresponding URL can be chosen using a key.

            The default browser is {default}. You can override the default
            browser on a per-account basis by adding an attribute named
            'browser' to the account.  An example of when you would specify the
            browser in an account would be an account associated with Tor hidden
            service, which generally can only be accessed using torbrowser:

                class SilkRoad(Account):
                    passcode = Passphrase()
                    username = Passphrase(length=2, sep='-')
                    url = 'http://silkroad6ownowfk.onion'
                    browser = 't'

            The available browsers are:
            {browsers}
        """).strip()
        return text.format(
            title=title(cls.DESCRIPTION), usage=cls.USAGE,
            default=default_browser, browsers=browsers
        )

    @classmethod
    def run(cls, command, args):
        # read command line
        cmdline = docopt(cls.USAGE, argv=[command] + args)

        # run the generator
        generator = PasswordGenerator(check_integrity=False)

        # determine the account and open the URL
        account = generator.get_account(cmdline['<account>'])
        account.open_browser(
            cmdline['<key>'], cmdline['--browser'], cmdline['--list']
        )


# Changed {{{1
class Changed(Command):
    NAMES = 'changed', 'C'
    DESCRIPTION = 'show changes since archive was created'
    USAGE = dedent("""
        Usage:
            avendesora changed
            avendesora C

        When you run the 'archive' command it overwrites the existing
        archive. If you have accidentally deleted an account or changed a
        secret, then replacing the archive could cause the last copy of the
        original information to be lost. To prevent this from occurring it
        is a good practice to run the 'changed' command before regenerating
        the archive.  It describes all of the changes that have occurred
        since the last time the archive was generated. You should only
        regenerate the archive once you have convinced yourself all of the
        changes are as expected.
    """).strip()

    @classmethod
    def run(cls, command, args):
        # define white space insensitive compare function:
        def differ(a, b):
            a = ' '.join(repr(a).replace(r'\n', ' ').split())
            b = ' '.join(repr(b).replace(r'\n', ' ').split())
            return a != b

        # read command line
        cmdline = docopt(cls.USAGE, argv=[command] + args)

        # read archive file
        archive_path = get_setting('archive_file')
        f = PythonFile(archive_path)
        archive = f.run()
        import arrow
        created = archive.get('CREATED')
        if created:
            created = arrow.get(created).format('YYYY-MM-DD hh:mm:ss A ZZ')
            output('archive created: %s' % created)
        archive_accounts = archive.get('ACCOUNTS')
        if not archive_accounts:
            raise PasswordError(
                'corrupt archive, ACCOUNTS missing.', culprit=archive_path
            )

        # run the generator
        generator = PasswordGenerator(warnings=False)

        # collect the accounts
        current_accounts = {}
        for account in generator.all_accounts():
            entry = account.archive()
            if entry:
                current_accounts[account.get_name()] = entry

        # determine which fields to ignore
        dynamic_fields = Collection(get_setting('dynamic_fields'))

        # report any new or missing accounts
        new = set(current_accounts.keys()) - set(archive_accounts.keys())
        missing = set(archive_accounts.keys()) - set(current_accounts.keys())
        for each in sorted(new):
            output('new account:', each)
        for each in sorted(missing):
            output('missing account:', each)

        # for the common accounts, report any differences in the fields
        common = set(archive_accounts.keys()) & set(current_accounts.keys())
        for account_name in sorted(common):
            archive_account = archive_accounts[account_name]
            current_account = current_accounts[account_name]

            # report any new or missing fields
            new = set(current_account.keys()) - set(archive_account.keys())
            missing = set(archive_account.keys()) - set(current_account.keys())
            for each in sorted(new):
                output(account_name, 'new field', each, sep=': ')
            for each in sorted(missing):
                output(account_name, 'missing field', each, sep=': ')

            # for the common fields, report any differences in the values
            shared = set(archive_account.keys()) & set(current_account.keys())
            for field_name in sorted(shared):
                if field_name in dynamic_fields:
                    continue
                try:
                    archive_value = archive_account[field_name]
                    current_value = current_account[field_name]
                    if is_collection(current_value) != is_collection(archive_value):
                        output(account_name, 'field dimension differs', field_name, sep=': ')
                    elif is_collection(current_value):
                        archive_items = Collection(archive_account[field_name]).items()
                        current_items = Collection(current_account[field_name]).items()
                        archive_keys = set(k for k, v in archive_items)
                        current_keys = set(k for k, v in current_items)
                        new = current_keys - archive_keys
                        missing = archive_keys - current_keys
                        for each in sorted(new):
                            output(account_name, field_name, 'new member', each, sep=': ')
                        for each in sorted(missing):
                            output(account_name, field_name, 'missing member', each, sep=': ')
                        for k in sorted(archive_keys & current_keys):
                            if differ(archive_value[k], current_value[k]):
                                output(account_name, 'member differs', '%s[%s]' % (field_name, k), sep=': ')
                    else:
                        if differ(archive_value, current_value):
                            output(account_name, 'field differs', field_name, sep=': ')
                except Exception:
                    error(
                        'unanticipated situation.',
                        culprit=(account_name, field_name)
                    )
                    raise


# Conceal {{{1
class Conceal(Command):
    NAMES = 'conceal', 'c'
    DESCRIPTION = 'conceal text by encoding it'
    USAGE = dedent("""
        Usage:
            avendesora conceal [options] [<text>]
            avendesora c       [options] [<text>]

        Options:
            -e <encoding>, --encoding <encoding>
                                    Encoding used when concealing information.
            -g <id>, --gpg-id <id>  Use this ID when creating any missing encrypted files.
                                    Use commas with no spaces to separate multiple IDs.
            -h <path>, --gpg-home <path>
                                    GPG home directory (default is ~/.gnupg).
            -s, --symmetric         Encrypt with a passphrase rather than using your
                                    GPG key (only appropriate for gpg encodings).
    """).strip()

    @classmethod
    def help(cls):
        encodings = []
        for name, desc in ObscuredSecret.encodings():
            encodings.append('%s:\n%s' % (name, indent(desc, '    ')))
        encodings = '\n\n'.join(encodings)
        text = dedent("""
            {title}

            {usage}

            Possible encodings include{default_encoding}:

            {encodings}

            Though available as an option for convenience, you should not pass
            the text to be hidden as an argument as it is possible for others to
            examine the commands you run and their argument list. For any
            sensitive secret, you should simply run 'avendesora conceal' and
            then enter the secret text when prompted.
        """).strip()
        default_encoding = ObscuredSecret.default_encoding()
        return text.format(
            title=title(cls.DESCRIPTION), usage=cls.USAGE, encodings=encodings,
            default_encoding=' (default encoding is %s)' % default_encoding,
        )

    @classmethod
    def run(cls, command, args):
        # read command line
        cmdline = docopt(cls.USAGE, argv=[command] + args)

        # get the text
        encoding = cmdline['--encoding']
        text = cmdline['<text>']
        symmetric = cmdline['--symmetric']

        # get the gpg ids of those who will be able to decrypt the file
        gpg_ids = cmdline['--gpg-id']
        if gpg_ids:
            gpg_ids = gpg_ids.split(',')
        else:
            gpg_ids = get_setting('gpg_ids', [])

        if cmdline['--gpg-home']:
            override_setting('gpg_home', cmdline['--gpg-home'])

        if not text:
            output('Enter text to obscure, type ctrl-d to terminate.')
            text = sys.stdin.read()[:-1]

        # transform and output the string
        output(ObscuredSecret.hide(text, encoding, True, symmetric, gpg_ids))


# Edit {{{1
class Edit(Command):
    NAMES = 'edit', 'e'
    DESCRIPTION = 'edit an account'
    USAGE = dedent("""
        Usage:
            avendesora edit <account>
            avendesora e    <account>
    """).strip()

    @classmethod
    def help(cls):
        text = dedent("""
            {title}

            {usage}

            Opens an existing account in your editor.

            You can specify the editor by changing the 'edit_account' setting in
            the config file (~/.config/avendesora/config).
        """).strip()
        return text.format(title=title(cls.DESCRIPTION), usage=cls.USAGE)

    @classmethod
    def run(cls, command, args):
        # read command line
        cmdline = docopt(cls.USAGE, argv=[command] + args)

        # run the generator
        generator = PasswordGenerator()

        # determine the account file and back it up
        try:
            account = generator.get_account(cmdline['<account>'])
            accounts_file = PythonFile(account._file_info.path)
            accounts_file.backup('.saved')
            account_name = account.__name__
        except OSErrors as e:
            raise PasswordError(os_error(e))

        # allow the user to edit, and then check and make sure it is valid
        try:
            while True:
                GenericEditor.run(accounts_file.path, account_name)

                # check the changes to see if there are any issues
                try:
                    accounts_file.run()
                    break
                except PasswordError as e:
                    error(e)
                    response = query_user('Try again?')
                    if response.lower() not in ['y', 'yes']:
                        raise PasswordError(
                            'Giving up, restoring original version.'
                        )
            accounts_file.chmod()

        except PasswordError:
            accounts_file.restore()
            raise
        except OSErrors as e:
            accounts_file.restore()
            raise PasswordError(os_error(e))


# Find {{{1
class Find(Command):
    NAMES = 'find', 'f'
    DESCRIPTION = 'find an account'
    USAGE = dedent("""
        Find accounts whose name contains the search text.

        Usage:
            avendesora find <text>
            avendesora f    <text>
    """).strip()

    @classmethod
    def run(cls, command, args):
        # read command line
        cmdline = docopt(cls.USAGE, argv=[command] + args)

        # run the generator
        generator = PasswordGenerator(check_integrity=False)

        # find accounts whose name matches the criteria
        to_print = []
        for acct in generator.find_accounts(cmdline['<text>']):
            aliases = Collection(getattr(acct, 'aliases', []))
            desc = getattr(acct, 'desc', None)
            to_print.append(
                join(acct.get_name(), aliases=aliases, desc=desc, template=(
                    '{} ({aliases:|, }) -- {desc}',
                    '{} ({aliases:|, })',
                    '{} -- {desc}',
                    '{}'
                ))
            )
        output(cmdline['<text>']+ ':')
        output('    ' + ('\n    '.join(sorted(to_print))))


# Help {{{1
class Help(Command):
    NAMES = 'help', 'h'
    DESCRIPTION = 'give information about commands or other topics'
    USAGE = dedent("""
        Usage:
            avendesora help [options] [<topic>]
            avendesora h    [options] [<topic>]

        Options:
            -s, --search            list topics that include <topic> as a search term.
            -b, --browse            open the topic in your default browser.
    """).strip()

    @classmethod
    def run(cls, command, args):
        # read command line
        cmdline = docopt(cls.USAGE, argv=[command] + args)

        # in this case we don't need to bother with a possibly encrypted logfile
        override_setting('discard_logfile', True)

        from .help import HelpMessage
        HelpMessage.show(cmdline['<topic>'], cmdline['--browse'], cmdline['--search'])


# Identity {{{1
class Identity(Command):
    NAMES = 'identity', 'ident', 'i'
    DESCRIPTION = 'generate an identifying response to a challenge'
    USAGE = dedent("""
        Usage:
            avendesora identity [<name> [<challenge>...]]
            avendesora ident    [<name> [<challenge>...]]
            avendesora i        [<name> [<challenge>...]]
    """).strip()

    @classmethod
    def help(cls):
        text = dedent("""
            {title}

            {usage}

            This command allows you to generate a response to any challenge.
            The response identifies you to a partner with whom you have shared
            an account.

            If you run the command with no arguments, it prints the list of
            valid names. If you run it with no challenge, one is created for you
            based on the current time and date.

            If you have a remote partner to whom you wish to prove your
            identity, have that partner use avendesora to generate a challenge
            and a response based on your shared secret. Then the remote partner
            provides you with the challenge and you run avendesora with that
            challenge to generate the same response, which you provide to your
            remote partner to prove your identity.

            You are free to explicitly specify a challenge to start the process,
            but it is important that it be unpredictable and that you not use
            the same challenge twice. As such, it is recommended that you not
            provide the challenge. In this situation, one is generated for you
            based on the time and date.

            Consider an example that illustrates the process. In this example,
            Siuan is confirming the identity of Moiraine, where both Siuan and
            Moiraine are assumed to have shared *Avendesora* accounts.  Siuan
            runs *Avendesora* as follows and remembers the response::

                > avendesora identity moiraine
                challenge: slouch emirate bedeck brooding
                response: spear disable local marigold

            This assumes that moiraine is the name, with any extension removed,
            of the file that Siuan uses to contain their shared accounts.

            Siuan communicates the challenge to Moiraine but not the response.
            Moiraine then runs *Avendesora* with the given challenge::

                > avendesora identity siuan slouch emirate bedeck brooding
                challenge: slouch emirate bedeck brooding
                response: spear disable local marigold

            In this example, siuan is the name of the file that Moiraine uses to
            contain their shared accounts.

            To complete the process, Moiraine returns the response to Siuan, who
            compares it to the response she received to confirm Moiraine's
            identity.  If Siuan has forgotten the desired response, she can also
            specify the challenge to the :ref:`identity command <identity
            command>` to regenerate the expected response.

            Alternately, when Siuan sends a message to Moiraine, she can
            proactively prove her identity by providing both the challenge and
            the response. Moiraine could then run the *identity* command with
            the challenge and confirm that she gets the same response. Other
            than herself, only Siuan could predict the correct response to any
            challenge.  However, this is not recommended as it would allow
            someone with brief access to Suian's Avendesora, perhaps Leane her
            Keeper, to generate and store multiple challenge/response pairs.
            Leane could then send messages to Moiraine while pretending to be
            Siuan using the saved challenge/response pairs.  The subterfuge
            would not work if Moiraine generated the challenge unless Leane
            currently has access to Siuan's Avendesora.
        """).strip()
        return text.format(title=title(cls.DESCRIPTION), usage=cls.USAGE)

    @classmethod
    def run(cls, command, args):
        # read command line
        cmdline = docopt(cls.USAGE, argv=[command] + args)

        # run the generator
        generator = PasswordGenerator()
        generator.account_files.load_account_files()

        if cmdline['<name>']:
            try:
                name = cmdline['<name>']
                challenge = ' '.join(cmdline['<challenge>'])
                c, r = generator.challenge_response(name, challenge)
                output(c, culprit='challenge')
                output(r, culprit='response')
            except PasswordError as e:
                e.report()
        else:
            names = sorted(generator.shared_secrets.keys())
            output('Available names:')
            for name in names:
                output('  ', name)


# Initialize {{{1
class Initialize(Command):
    NAMES = 'initialize', 'init', 'I'
    DESCRIPTION = 'create initial set of Avendesora files'
    USAGE = dedent("""
        Usage:
            avendesora initialize [options]
            avendesora init       [options]
            avendesora I          [options]

        Options:
            -g <id>, --gpg-id <id>  Use this ID when creating any missing encrypted files.
                                    Use commas with no spaces to separate multiple IDs.
            -h <path>, --gpg-home <path>
                                    GPG home directory (default is ~/.gnupg).
    """).strip()

    @classmethod
    def help(cls):
        text = dedent("""
            {title}

            {usage}

            Create Avendesora data directory (~/.config/avendesora) and populate
            it with initial versions of all essential files.

            It is safe to run this command even after the data directory and
            files have been created. Doing so will simply recreate any missing
            files.  Existing files are not modified.
        """).strip()
        return text.format(title=title(cls.DESCRIPTION), usage=cls.USAGE)

    @classmethod
    def run(cls, command, args):
        # read command line
        cmdline = docopt(cls.USAGE, argv=[command] + args)
        if cmdline['--gpg-home']:
            override_setting('gpg_home', cmdline['--gpg-home'])

        # get the gpg ids of those who will be able to decrypt the file
        gpg_ids = cmdline['--gpg-id']
        if gpg_ids:
            gpg_ids = gpg_ids.split(',')
            # save the gpg_ids for the logfile in case it is encrypted.
            if not get_setting('gpg_ids'):
                override_setting('gpg_ids', gpg_ids)
        else:
            gpg_ids = get_setting('gpg_ids', [])

        # run the generator
        generator = PasswordGenerator(init=True, gpg_ids=gpg_ids)


# Log {{{1
class Log(Command):
    NAMES = 'log',
    DESCRIPTION = 'open the logfile'
    USAGE = dedent("""
        Usage:
            avendesora [options] log

        Options:
            -d, --delete   Delete the logfile rather than opening it.
    """).strip()

    @classmethod
    def help(cls):
        text = dedent("""
            {title}

            {usage}

            Opens the logfile in your editor.

            You can specify the editor by changing the 'edit_account' setting in
            the config file (~/.config/avendesora/config).
        """).strip()
        return text.format(title=title(cls.DESCRIPTION), usage=cls.USAGE)

    @classmethod
    def run(cls, command, args):
        # read command line
        cmdline = docopt(cls.USAGE, argv=[command] + args)

        # don't clobber the existing logfile
        override_setting('discard_logfile', True)

        # open the logfile
        logfile = get_setting('log_file')

        if cmdline['--delete']:
            try:
                rm(logfile)
                return
            except OSErrors as e:
                raise PasswordError(os_error(e))
        if not logfile.exists():
            raise PasswordError('log file was not found.')
        else:
            GenericEditor.run(logfile)


# Login Credentials {{{1
class LoginCredentials(Command):
    NAMES = 'credentials', 'login', 'l'
    DESCRIPTION = 'show login credentials'
    USAGE = dedent("""
        Displays the account's login credentials, which generally consist of an
        identifier and a secret.

        Usage:
            avendesora credentials [options] <account>
            avendesora login       [options] <account>
            avendesora l           [options] <account>

        Options:
            -S, --seed              Interactively request additional seed for
                                    generated secrets.
    """).strip()


    @classmethod
    def help(cls):
        idents=conjoin(get_setting('credential_ids').split(), conj=' or ')
        secrets=conjoin(get_setting('credential_secrets').split(), conj=' or ')
        text = dedent("""
            {title}

            {usage}

            The credentials can be specified explicitly using the credentials
            setting in your account. For example::

                credentials = 'usernames.0 usernames.1 passcode'

            If credentials is not specified then the first of the following will
            be used if available:

                id: {idents}
                secret: {secrets}

            If your credentials include more than one secret they will be
            presented one at a time for one minute each. You can cut the minute
            short by typing Ctrl-C.
        """).strip()
        return text.format(
            title=title(cls.DESCRIPTION), usage=cls.USAGE,
            idents=idents, secrets=secrets,
        ).strip()

    @classmethod
    def run(cls, command, args):
        # read command line
        cmdline = docopt(cls.USAGE, argv=[command] + args)

        # run the generator
        generator = PasswordGenerator(check_integrity=False)

        # determine the account and output specified information
        account_name = cmdline['<account>']
        writer = get_writer(tty=True)
        account = generator.get_account(account_name, cmdline['--seed'])
        credentials = account.get_scalar('credentials', default=None)
        if credentials:
            credentials = credentials.split()
        else:
            ids = Collection(get_setting('credential_ids'))
            secrets = Collection(get_setting('credential_secrets'))
            for each in ids:
                if account.has_field(each):
                    identity = each
                    break
            else:
                identity = None
            for each in secrets:
                if account.has_field(each):
                    secret = each
                    break
            else:
                secret = None
            if not identity and not secret:
                raise PasswordError('credentials not found.')
            credentials = [identity, secret]
        for each in cull(credentials):
            writer.display_field(account, each)


# New {{{1
class New(Command):
    NAMES = 'new', 'N'
    DESCRIPTION = 'create new accounts file'
    USAGE = dedent("""
        Usage:
            avendesora new [options] <name>
            avendesora N   [options] <name>

        Options:
            -g <id>, --gpg-id <id>  Use this ID when creating any missing encrypted files.
                                    Use commas with no spaces to separate multiple IDs.
    """).strip()

    @classmethod
    def help(cls):
        text = dedent("""
            {title}

            {usage}

            Creates a new accounts file. Accounts that share the same file share
            the same master seed by default and, if the file is encrypted,
            can be decrypted by the same recipients.

            Generally you create new accounts files for each person or group
            with which you wish to share accounts. You also use separate files
            for passwords with different security domains. For example, a
            high-value passwords might be placed in an encrypted file that would
            only be placed highly on protected computers. Conversely, low-value
            passwords might be contained in perhaps an unencrypted file that is
            found on many computers.

            Add a '.gpg' extension to <name> to encrypt the file.
        """).strip()
        return text.format(title=title(cls.DESCRIPTION), usage=cls.USAGE)

    @classmethod
    def run(cls, command, args):
        # read command line
        cmdline = docopt(cls.USAGE, argv=[command] + args)

        # get the gpg ids of those who will be able to decrypt the file
        gpg_ids = cmdline['--gpg-id']
        if gpg_ids:
            gpg_ids = gpg_ids.split(',')
        else:
            gpg_ids = get_setting('gpg_ids', [])

        # run the generator
        generator = PasswordGenerator(
            init=cmdline['<name>'], gpg_ids=sorted(gpg_ids)
        )


# Phonetic Alphabet {{{1
class PhoneticAlphabet(Command):
    NAMES = 'phonetic', 'alphabet', 'p'
    DESCRIPTION = 'display NATO phonetic alphabet'
    USAGE = dedent("""
        Usage:
            avendesora alphabet [<text>]
            avendesora phonetic [<text>]
            avendesora p [<text>]
    """).strip()

    @classmethod
    def help(cls):
        text = dedent("""
            {title}

            {usage}

            If <text> is given, it is converted character by character to the
            phonetic alphabet. If not given, the entire phonetic alphabet is
            displayed.
        """).strip()
        return text.format(title=title(cls.DESCRIPTION), usage=cls.USAGE)

    @classmethod
    def run(cls, command, args):
        words = """
            Alfa Bravo Charlie Delta Echo Foxtrot Golf Hotel India Juliett Kilo
            Lima Mike November Oscar Papa Quebec Romeo Sierra Tango Uniform
            Victor Whiskey X-ray Yankee Zulu
        """.split()
        mapping = {w[0].lower():w for w in words}
        mapping.update({
            '0':'Zero', '1':'One', '2':'Two', '3':'Three', '4':'Four',
            '5':'Five', '6':'Six', '7':'Seven', '8':'Eight', '9':'Nine'
        })

        # read command line
        cmdline = docopt(cls.USAGE, argv=[command] + args)
        arg = cmdline['<text>']

        if arg:
            converted = []
            for c in arg.lower():
                converted.append(mapping.get(c, c))
            output(' '.join(converted).lower())
        else:
            output('Phonetic alphabet:')
            output(columns(words))


# Questions {{{1
class Questions(Command):
    NAMES = 'questions', 'quest', 'q', 'qc'
    DESCRIPTION = 'answer a security question'
    USAGE = dedent("""
        Displays the security questions and then allows you to select one
        to be answered.

        Usage:
            avendesora questions [options] <account> [<field>]
            avendesora quest     [options] <account> [<field>]
            avendesora q         [options] <account> [<field>]
            avendesora qc        [options] <account> [<field>]

        Options:
            -c, --clipboard         Write output to clipboard rather than stdout.
            -S, --seed              Interactively request additional seed for
                                    generated secrets.

        The 'qc' command is a shortcut for 'questions --clipboard'.

        Request the answer to a security question by giving the account name to
        this command.  For example:

            avendesora questions bank

        It will print out the security questions for *bank* account along with
        an index. Specify the index of the question you want answered.

        By default *Avendesora* looks for the security questions in the
        *questions* field.  If your questions are in a different field, just
        specify the name of the field on the command line:

            avendesora questions bank verbal
    """).strip()

    @classmethod
    def run(cls, command, args):
        # read command line
        cmdline = docopt(cls.USAGE, argv=[command] + args)
        use_clipboard = cmdline['--clipboard'] or cmdline['qc']

        # run the generator
        generator = PasswordGenerator(check_integrity=False)

        # get the questions
        account_name = cmdline['<account>']
        account = generator.get_account(account_name, cmdline['--seed'])
        field = cmdline.get('<field>')
        if not field:
            field = get_setting('default_vector_field')
        try:
            questions = getattr(account, field)
        except AttributeError:
            raise PasswordError('unknown field.', culprit=field)
        if not is_collection(questions):
            raise PasswordError('unknown field.', culprit=field)

        contains_secret = False
        for k, v in Collection(questions).items():
            try:
                value = v.get_description()
                contains_secret = True
            except AttributeError:
                value = v
            display(k, value, template='{}: {}')

        if contains_secret:
            response = query_user('Which question?')
            writer = get_writer(clipboard=use_clipboard, stdout=False)
            writer.display_field(account, field + '.' + response)


# Reveal {{{1
class Reveal(Command):
    NAMES = 'reveal', 'r'
    DESCRIPTION = 'reveal concealed text'
    USAGE = dedent("""
        Transform concealed text to reveal its original form.

        Usage:
            avendesora reveal [<text>]
            avendesora r      [<text>]

        Options:
            -e <encoding>, --encoding <encoding>
                                    Encoding used when revealing information.
    """).strip()

    @classmethod
    def help(cls):
        text = dedent("""
            {title}

            {usage}

            Though available as an option for convenience, you should not pass
            the text to be revealed as an argument as it is possible for others
            to examine the commands you run and their argument list. For any
            sensitive secret, you should simply run 'avendesora reveal' and then
            enter the encoded text when prompted.
        """).strip()
        return text.format(
            title=title(cls.DESCRIPTION), usage=cls.USAGE
        )

    @classmethod
    def run(cls, command, args):
        # read command line
        cmdline = docopt(cls.USAGE, argv=[command] + args)

        # get the text
        text = cmdline['<text>']
        if not text:
            output('Enter the obscured text, type ctrl-d to terminate.')
            text = sys.stdin.read()

        # transform and output the string
        output(ObscuredSecret.show(text))


# Search {{{1
class Search(Command):
    NAMES = 'search', 's'
    DESCRIPTION = 'search accounts'
    USAGE = dedent("""
        Search for accounts whose values contain the search text.

        Usage:
            avendesora search <text>
            avendesora s      <text>
    """).strip()

    @classmethod
    def run(cls, command, args):
        # read command line
        cmdline = docopt(cls.USAGE, argv=[command] + args)

        # run the generator
        generator = PasswordGenerator(check_integrity=False)

        # search for accounts that match search criteria
        to_print = []
        for acct in generator.search_accounts(cmdline['<text>']):
            aliases = Collection(getattr(acct, 'aliases', []))
            desc = getattr(acct, 'desc', None)
            to_print.append(
                join(acct.get_name(), aliases=aliases, desc=desc, template=(
                    '{} ({aliases:|, }) -- {desc}',
                    '{} ({aliases:|, })',
                    '{} -- {desc}',
                    '{}'
                ))
            )
        output(cmdline['<text>'] + ':')
        output('    ' + ('\n    '.join(sorted(to_print))))


# Value {{{1
class Value(Command):
    NAMES = 'value', 'val', 'v', 'vc'
    DESCRIPTION = 'show an account value'
    USAGE = dedent("""
        Produce an account value. If the value is secret, it is produced only
        temporarily unless --stdout is specified.

        Usage:
            avendesora value [options] [<account> [<field>]]
            avendesora val   [options] [<account> [<field>]]
            avendesora v     [options] [<account> [<field>]]
            avendesora vc    [options] [<account> [<field>]]

        Options:
            -c, --clipboard         Write output to clipboard rather than stdout.
            -s, --stdout            Write output to the standard output without
                                    any annotation or protections.
            -S, --seed              Interactively request additional seed for
                                    generated secrets.
            -v, --verbose           Add additional information to log file to
                                    help identify issues in account discovery.
            -T <title>, --title <title>
                                    Use account discovery on this title.

        The 'vc' command is a shortcut for 'value --clipboard'.

        You request a scalar value by specifying its name after the account.
        For example:

            avendesora value bank pin

        If the requested value is composite (an array or dictionary), you should
        also specify a key that indicates which of the composite values you
        want. For example, if the 'accounts' field is a dictionary, you specify
        accounts.checking or accounts[checking] to get information on your
        checking account. If the value is an array, you give the index of the
        desired value. For example, questions.0 or questions[0]. If you only
        specify a number, then the name is assumed to be 'questions', as in the
        list of security questions (this can be changed by specifying the
        desired name as the 'default_vector_field' in the account or the config
        file).

        The field may be also be a script, with is nothing but a string that it
        output as given except that embedded attributes are replaced by account
        field values. For example:

            avendesora value bank '{accounts.checking}: {passcode}'

        If no value is requested the result produced is determined by the value
        of the 'default' attribute (this can be changed by specifying
        'default_field' in the config file). If no value is given for 'default',
        then the *passcode*, *password*, or *passphrase* attribute is produced
        (this can be changed by specifying the 'default_field' in the account or
        the config file).  If 'default' is a script (see 'avendesora help
        scripts') then the script is executed. A typical script might be
        'username: {username}, password: {passcode}'. It is best if the script
        produces a one line output if it contains secrets. If not a script, the
        value of 'default' should be the name of another attribute, and the
        value of that attribute is shown.

        Normally the value command attempts to protects secrets. It does so
        clearing the screen after a minute. If multiple secrets are requested,
        you must either wait a minute to see each subsequent secret or type
        Ctrl-C to clear the current secret and move on.  If you use --clipboard,
        the clipboard is cleared after a minute.  However, if you use --stdout
        this clearing of the secret does not occur. The --stdout option is
        generally used with communicating with other Linux commands.  For
        example, you can send a passcode to the standard input of a command as
        follows:

            avendesora value --stdout gpg | gpg --passphrase-fd 0 ...

        You can place the username and password on a command line as follows:

            curl --user `avendesora value -s apache '{username}:{passcode}'` ...

        Be aware that it is possible for other users on shared Linux machines to
        see the command line arguments of your commands, so passing secrets as
        command arguments should only be used for low value secrets.

        If no account is requested, then Avendesora attempts to determine the
        appropriate account through discovery (see 'avendesora help discovery').
        Normally Avendesora is called in this manner from your window manager.
        You would arrange for it to be run when you type a hot key. In this case
        Avendesora determines which account to use from information available
        from the environment, information like the title on active window. In
        this mode, Avendesora mimics the keyboard when producing its output.

        The verbose and title options are used when debugging account
        discovery. The verbose option adds more information about the
        discovery process to the logfile (~/.config/avendesora/log.gpg). The
        title option allows you to override the active window title so you can
        debug title-based discovery. Specifying the title option also scrubs
        the output and outputs directly to the standard output rather than
        mimicking the keyboard so as to avoid exposing your secret.

        If the --stdout option is not specified, the value command still writes
        to the standard output if it is associated with a TTY (if Avendesora is
        outputting directly to a terminal). If standard output is not a TTY,
        Avendesora mimics the keyboard and types the desired value directly into
        the active window.  There are two common situations where standard
        output is not a TTY: when Avendesora is being run by your window manager
        in response to you pressing a hot key or when the output of Avendesora
        is fed into a pipeline.  In the second case, mimicking the keyboard is
        not what you want; you should use --stdout to assure the chosen value is
        sent to the pipeline as desired.
    """).strip()

    @classmethod
    def run(cls, command, args):
        # read command line
        cmdline = docopt(cls.USAGE, argv=[command] + args)
        use_clipboard = cmdline['--clipboard'] or cmdline['vc']

        # mute any warnings if using --stdout
        if cmdline['--stdout']:
            from inform import get_informer
            get_informer().suppress_output()

        # run the generator
        generator = PasswordGenerator(check_integrity=False)

        # determine the account and output specified information
        account_name = cmdline['<account>']
        if account_name:
            account = generator.get_account(account_name, cmdline['--seed'])
            writer = get_writer(
                clipboard=use_clipboard, stdout=cmdline['--stdout']
            )
            writer.display_field(account, cmdline['<field>'])
        else:
            # use discovery to determine account
            script = generator.discover_account(
                title=cmdline['--title'], verbose=cmdline['--verbose']
            )
            writer = get_writer(tty=False)
            writer.run_script(script, dryrun=bool(cmdline['--title']))


# Values {{{1
class Values(Command):
    NAMES = 'values', 'vals', 'V'
    DESCRIPTION = 'display all account values'
    USAGE = dedent("""
        Show all account values.

        Usage:
            avendesora values <account>
            avendesora vals   <account>
            avendesora V      <account>
    """).strip()

    @classmethod
    def run(cls, command, args):
        # read command line
        cmdline = docopt(cls.USAGE, argv=[command] + args)

        # run the generator
        generator = PasswordGenerator(check_integrity=False)

        # determine the account
        account = generator.get_account(cmdline['<account>'])
        account.write_summary()


# Version {{{1
class Version(Command):
    NAMES = 'version',
    DESCRIPTION = 'display Avendesora version'
    USAGE = dedent("""
        Usage:
            avendesora version
    """).strip()

    @classmethod
    def run(cls, command, args):
        # read command line
        cmdline = docopt(cls.USAGE, argv=[command] + args)

        # get the Python version
        python = 'Python %s.%s.%s' % (
            sys.version_info.major,
            sys.version_info.minor,
            sys.version_info.micro,
        )

        # output the Avendesora version along with the Python version
        from .__init__ import __version__, __released__
        output('Avendesora version: %s (%s) [%s].' % (
            __version__, __released__, python
        ))




# Avendesora Password Generator Preferences

# License {{{1
# Copyright (C) 2016-18 Kenneth S. Kundert
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses/.

# Imports {{{1
from textwrap import dedent
import re
from appdirs import user_config_dir

# Non-Config Settings {{{1
# These value can be accessed with get_settings,
# but should not be found in the config file
NONCONFIG_SETTINGS = {
    'config_file': 'config',
    'config_doc_file': 'config.doc',
    'settings_dir': user_config_dir('avendesora'),
    'default_accounts_file': 'accounts.gpg',
    'default_stealth_accounts_file': 'stealth_accounts',
    'charsets_hash': '0120b695fb247d4993b1595ab087cc15',
    'dict_hash': '11fe5bc734f4a956c37d7cb3da16ab3f',
    'secrets_hash': '2fdd6fd957c1b9d511ee40f0881bf3b3',
    'discard_logfile': False,
    'commonly_mistaken_attributes': {
        'url': 'urls',
        'alias': 'aliases',
        'description': 'desc',
    }
}


# Config Settings {{{1
# These are the default values for settings that may be found in the config file
CONFIG_DEFAULTS = {
    'accounts_files': [
        NONCONFIG_SETTINGS['default_accounts_file'],
        NONCONFIG_SETTINGS['default_stealth_accounts_file'],
    ],
    'account_list_file': 'accounts_files',
    'archive_file': 'archive.gpg',
    'archive_stale': 1,
    'previous_archive_file': 'archive.prev.gpg',
    'config_dir_mask': 0o077,
    'account_file_mask': 0o077,
    'credential_ids': 'username email',
    'credential_secrets': 'passcode password passphrase',
    'label_color': 'blue',
    'highlight_color': 'magenta',
    'color_scheme': 'dark',
    'browsers': {
        'c':  'google-chrome {url}',
        'ci': 'google-chrome --incognito {url}',
        'f':  'firefox -new-tab {url}',
        'fp':  'firefox -private-window {url}',
        'q':  'qutebrowser {url}',
        't':  'torbrowser {url}',
        'x':  'xdg-open {url}', # system default browser
    },
    'help_url': 'https://avendesora.readthedocs.io/en/latest',
    'default_field': 'passcode password passphrase',
    'default_vector_field': 'questions',
    'dynamic_fields': '',
    'default_browser': 'x',
    'display_time': 60,
    'dictionary_file': None,
    'encoding': 'utf-8',
    'edit_account': (
        'gvim',                     # use gvim -v so that user can access
        '-v',                       # the X clipboard buffers
        '+silent! /^class {account}(Account):/',
        '+silent! normal zozt',     # open the fold, position near top of screen
        '{filepath}'
    ),
    'edit_template': (
        'gvim',                     # use gvim -v so that user can access
        '-v',                       # the X clipboard buffers
        '+silent! /_[A-Z0-9_]\+_/', # matches user modifiable template fields
                                    # fields take the form '_AAA_'
        '+silent! normal zozt',     # open the fold, position near top of screen
        '{filepath}'
    ),
    'hashes_file': 'hashes',
    'indent': '    ',
    'log_file': 'log.gpg',
    'default_protocol': 'https',
    'user_key_file': 'key.gpg',
    'use_pager': True,
    'selection_utility': 'gtk',
    'verbose': False,
    'default_account_template': 'bank',
    'account_templates': {
        'website': dedent("""
            class _NAME_(Account): # %s1
                desc = '_DESCRIPTION_'
                aliases = '_ALIAS1_ _ALIAS2_'
                username = '_USERNAME_'
                email = '_EMAIL_'
                passcode = PasswordRecipe('12 2u 2d 2s')
            # Avendesora: length is 12, includes 2 upper, 2 digits and 2 symbols
            # Avendesora: Alternatively use '12 2u 2d 2c!@#$&' to specify valid symbol characters.
            # Avendesora: Alternatively use Passphrase()
                questions = [
                    Question("_QUESTION1_?"),
                    Question("_QUESTION2_?"),
                    Question("_QUESTION3_?"),
                ]
                urls = '_URL_'
            # Avendesora: specify URL or URLs to be used by the browse command.
            # Avendesora: generally this is the URL of the sign-in page.
                discovery = RecognizeURL(
                    'https://_URL_',
                    script='{email}{tab}{passcode}{return}'
                )
            # Avendesora: Specify multiple urls to recognizer if several pages need same script.
            # Avendesora: Specify list of recognizers if multiple pages need different scripts.

            # Avendesora: Tailor the account entry to suit you needs.
            # Avendesora: You can add or delete class attributes as you see fit.
            # Avendesora: The 'n' key should take you to the next field name.
            # Avendesora: Use 'cw' to specify a field value, or 'dw' to delete it if unneeded.
            # Avendesora: Fields surrounded by << and >> will be hidden.
            # Avendesora: All lines that begin with '# Avendesora:' are removed.
        """ % (3*'{')),
        'shell': dedent("""
            class _NAME_(Account): # %s1
                desc = '_DESCRIPTION_'
                aliases = '_ALIAS1_ _ALIAS2_'
                passcode = Passphrase()
            # Avendesora: Alternatively use PasswordRecipe('12 2u 2d 2s')
            # Avendesora: or '12 2u 2d 2c!@#$&' to specify valid symbol characters.
                discovery = RecognizeTitle(
                    '_TITLE1_', '_TITLE2_',
                    script='{passcode}{return}'
                )

            # Avendesora: Tailor the account entry to suit you needs.
            # Avendesora: You can add or delete class attributes as you see fit.
            # Avendesora: The 'n' key should take you to the next field name.
            # Avendesora: Use 'cw' to specify a field name, or delete it if unneeded.
            # Avendesora: Fields surrounded by << and >> will be hidden.
            # Avendesora: All lines that begin with '# Avendesora:' are deleted.
        """ % (3*'{')),
        'bank': dedent("""
            class _NAME_(Account): # %s1
                desc = '_DESCRIPTION_'
                aliases = '_ALIAS1_ _ALIAS2_'
                username = '_NAME_'
                email = '_EMAIL_'
                accounts = {
                    'checking':   <<_ACCOUNT_NUMBER_>>,
                    'savings':    <<_ACCOUNT_NUMBER_>>,
                    'creditcard': <<_ACCOUNT_NUMBER_>>,
                }
                customer_service = '_PHONE_NUMBER_'
                passcode = PasswordRecipe('12 2u 2d 2s')
            # Avendesora: length is 12, includes 2 upper, 2 digits and 2 symbols
            # Avendesora: Alternatively use '12 2u 2d 2c!@#$&' to specify valid symbol characters.
            # Avendesora: Alternatively use Passphrase()
                verbal = Passphrase(length=2)
                pin = PIN()
                questions = [
                    Question("_QUESTION1_?"),
                    Question("_QUESTION2_?"),
                    Question("_QUESTION3_?"),
                ]
                urls = '_URL_'
            # Avendesora: specify urls to indicate the url to be used by the browse command.
                discovery = RecognizeURL(
                    'https://_URL_',
                    script='{email}{tab}{passcode}{return}'
                )
            # Avendesora: Specify multiple urls to recognizer if several pages need same script.
            # Avendesora: Specify list of recognizers if multiple pages need different scripts.

            # Avendesora: Tailor the account entry to suit you needs.
            # Avendesora: You can add or delete class attributes as you see fit.
            # Avendesora: The 'n' key should take you to the next field name.
            # Avendesora: Use 'cw' to specify a field name, or delete it if unneeded.
            # Avendesora: Fields surrounded by << and >> will be hidden.
            # Avendesora: All lines that begin with '# Avendesora:' are deleted.
        """ % (3*'{')),
    },

    # use absolute paths for executables so they cannot be maliciously replaced
    # by changing the path.
    'gpg_executable': '/usr/bin/gpg2',
    'gpg_home': '~/.gnupg',
    'gpg_armor': 'extension',
    'gpg_ids': None,
    'xdotool_executable': '/usr/bin/xdotool',
    'xsel_executable': '/usr/bin/xsel',
    'dmenu_executable': '/usr/bin/dmenu',
    'arp_executable': '/sbin/arp',
}

# The following could be config settings, but they do not seem worth promoting.
INITIAL_AUTOTYPE_DELAY = 0.25
    # Give the window manager a bit of time to close the 'choose' dialog and
    # refocus on the original window. Delay given in seconds.

# Initial config file {{{1
CONFIG_FILE_INITIAL_CONTENTS = dedent('''\
    # Avendesora Configuration
    #
    # This file is read by Avendesora. Use it to change any configuration
    # settings from their default values.  See {config_doc_file} for the list of
    # available settings along with a description and the default value of the
    # setting.

    # Default identity to use when creating encrypted files.
    gpg_ids = {gpg_ids}

    # vim: filetype=python sw=4 sts=4 et ai ff=unix nofen fileencoding={encoding} :
''')

# Initial config.doc file {{{1
CONFIG_DOC_FILE_INITIAL_CONTENTS = dedent('''\
    # Avendesora Configuration Documentation
    #
    # This file is generated and updated by Avendesora using the initialize
    # command. It documents the supported configuration settings and gives the
    # default values for those settings. However it is not read by Avendesora. To
    # change your configuration settings, you should modify your config file:
    # {config_file}. In general you want to include only those settings that
    # differ from their defaults in {config_file}. In this way, if the defaults
    # change as Avendesora is upgraded you will naturally pick up the new
    # values.
    #
    # This file is replaced every time you run 'avendesora initialize'. Any
    # changes you make to this file will be lost.

    # Avendesora settings
    log_file = {log_file}
        # The desired location of the log file (relative to config directory).
        # Adding a suffix of .gpg or .asc causes the file to be encrypted
        # (otherwise it can leak account names). Use None to disable logging.

    archive_file = {archive_file}
    previous_archive_file = {previous_archive_file}
        # The desired location of the archive files (relative to config director).
        # End the path in .gpg or .asc. Use None to disable archiving.

    archive_stale = {archive_stale}
        # The archive file is consider stale if it is this many days older than
        # the most recently updated account file.

    default_field = {default_field}
        # The name of the field to use for the value command when one is not
        # given. May be a space separated list of names, in which case the first
        # that is found is used.

    default_vector_field = {default_vector_field}
        # The name of the field to use when an integer is given as the argument
        # to the value command. In this case the field is expected to be a list
        # and the argument is taken to be the index of the desired value.
        # For example, if default_vector_field is 'question' and the argument
        # given with the value command is 1, then question[1] is produced.

    dynamic_fields = {dynamic_fields}
        # Fields whose values can change in real time. These fields will not be
        # mentioned by the *changed* command, even if their value differs from
        # when the last archive was created.

    credential_ids = {credential_ids}
        # A string that contains the field names (space separated) that should be
        # considered by the credentials command for the account identity.

    credential_secrets = {credential_secrets}
        # A string that contains the field names (space separated) that should be
        # considered by the credentials command for the primary account secret.

    display_time = {display_time}
        # The number of seconds that the secret will be displayed before it is
        # erase when writing to the TTY or clibboard.

    encoding = {encoding!r}
        # The unicode encoding to use when reading or writing files.

    edit_account = {edit_account}
        # The command used when editing an account. The command is given as
        # list of strings. The strings may contain {{filepath}} and {{account}},
        # which are replaced by the path to the file and the name of the
        # account.

    edit_template = {edit_template}
        # The command used when creating a new account that has been initialized
        # with a template. The command is given as list of strings. The strings
        # may contain {{filepath}}, which is replaced by the path to the file.

    browsers = {browsers}
        # A dictionary containing the supported browsers. For each entry the key
        # is the name to be used for the browser, and the value is string that
        # contains the command that invokes the browser. The value may contain
        # {{url}}, which is replaced by the URL to open.

    default_browser = {default_browser}
        # The name of the default browser. This name should be one of the keys
        # in the browsers dictionary.

    default_protocol = {default_protocol}
        # The default protocol to use for a URL if the protocol is not specified
        # in the requested URL. Generally this should be 'https' or 'http',
        # though 'https' is recommended.

    config_dir_mask = {config_dir_mask}
        # An integer that determines if a warning should be printed about the
        # file permissions on the Avendesora configuration directory
        # (~/.config/avendesora) being too loose. A bitwise and operation is
        # performed between this value and the actual file permissions, and if
        # the result is nonzero, a warning is printed.  Set to 0o000 to disable
        # the warning. Set to 0o077 to generate a warning if the configuration
        # directory is readable or writable by the group or others. Set to 0o007
        # to generated a warning if the directory is readable or writable by
        # others.

    account_file_mask = {account_file_mask}
        # An integer that determines if a warning should be printed about the
        # file permissions of the accounts and archive files being too loose.  A
        # bitwise and operation is performed between this value and the actual
        # file permissions, and if the result is nonzero, a warning is printed.
        # Set to 0o000 to disable the warning. Set to 0o077 to generate a
        # warning if the configuration directory is readable or writable by the
        # group or others. Set to 0o007 to generated a warning if the directory
        # is readable or writable by others.

    label_color = {label_color}
        # The color of the label use by the value and values commands.
        # Choose from black, red, green, yellow, blue, magenta, cyan, white.

    highlight_color = {highlight_color}
        # The color of the highlight use by the value and values commands.
        # Choose from black, red, green, yellow, blue, magenta, cyan, white.

    color_scheme = {color_scheme}
        # The color scheme used for the label color.  Choose from dark, light.
        # If the shell background color is light, use dark.

    use_pager = {use_pager}
        # Use a external program to break long output into pages.
        # May be either a boolean or a string. If a string the string is taken
        # to be a command line use to invoke a paging program (like 'more'). If
        # True, the program name is taken from the PAGER environment variable if
        # set, or 'less' is used if not set. If False, a paging program is not
        # used.

    selection_utility = {use_pager}
        Which utility should be used when it becomes necessary for you to
        interactively make a choice. Two utilities are available: 'gtk', the
        default, and 'dmenu'.

        'gtk' is the built-in selection. When needed it pops a small dialog box in
        the middle of the screen. You can use the 'j' and 'k' to navigate to your
        selection and 'l' to make the selection or 'h' to cancel.  Alternately you
        can use the arrow keys and Enter and Esc to navigate, select, and cancel.

        'dmenu' is an external utility, and must be installed. With *dmenu* you
        type the first few letters of your selection to highlight it, then type
        'Enter' to select or 'Esc' to cancel.

    verbose = {verbose}
        # Set this to True to generate additional information in the log file
        # that can help debug account discovery issues.  Normally it should be
        # False to avoid leaking account information into log file.
        # This is most useful when debugging account discovery, and in that case
        # this setting has largely been superseded by the use of the --title and
        # --verbose command line options.

    help_url = {help_url}
        # The base URL for Avendesora's online documentation.

    # Account templates
    # Use these settings to specify the available account templates that are
    # used when creating new accounts.  The templates are given as a dictionary
    # where the key is the name of the template and the value is the template
    # itself. The template is passed through textwrap.dedent() to remove any
    # leading white space.  Any lines that begin with '# Avendesora: ' represent
    # comments that can contain instructions to the user. They will are removed
    # when the account is created.
    account_templates = {account_templates}
    default_account_template = {default_account_template}

    # GPG settings
    # Information used by GPG when encrypting and decrypting files.
    gpg_ids = {gpg_ids}
        # the GPG IDs to use by default when creating encrypted files (the
        # archive and account files).
    gpg_armor = {gpg_armor}
        # In the GPG world, armoring a file means converting it to simple ASCI.
        # Choose between 'always', 'never' and 'extension' (.asc: armor, .gpg:
        # no).
    gpg_home = {gpg_home}
        # This is your GPG home directory. By default it will be ~/.gnupg.

    # External Executables
    # Use of full absolute paths for executables is recommended.
    gpg_executable = {gpg_executable}
        # Path to the gpg2 executable.
    xdotool_executable = {xdotool_executable}
    xsel_executable = {xsel_executable}
        # recommend '/usr/bin/xsel -p' if you wish to use mouse middle click
        # recommend '/usr/bin/xsel -b' if you wish to use mouse right click then paste
    dmenu_executable = {dmenu_executable}

    # vim: filetype=python sw=4 sts=4 et ai ff=unix nofen fileencoding={encoding} :
''')


# Initial hash file {{{1
HASH_FILE_INITIAL_CONTENTS = dedent('''\
    # Avendesora Hashes
    # vim: filetype=python sw=4 sts=4 et ai ff=unix fileencoding={encoding} :
    #
    # Changing the contents of the dictionary, secrets, or charsets files can change
    # the secrets you generate, thus you should be very careful in changing
    # these files.  These files may change when upgrading the program. Care is
    # taken to assure than your generated passwords do not change as a result of
    # an upgrade, but mistakes can be made.  Hashes for the contents of these
    # files are stored here, and the hashes are checked each time Avendesora is
    # run.  If you get a message indicating that the hashes have changed, it
    # does not mean that there is a problem. Rather it indicates that a change
    # has been made that could conceivably have created a problem and you
    # should do some checking before continuing to use the program. Once you are
    # convinced everything is working as expected, you should update these
    # hashes as needed.
    #
    # When upgrading the program, here is the procedure you should use.
    # 1. Before upgrading you should run 'avendesora changed' and examine all
    #    the differences and make sure they are all expected.
    # 2. Run 'avendesora archive' to create a clean archive.
    # 3. Update Avendesora
    #    This version may emit hash warnings.
    # 4. Run 'avendesora changed'. If no changes are noted, it is okay to update
    #    the appropriate hashes below.
    # 5. If there are changes, do not upgrade and report the situation to
    #    Avendesora's authors.

    charsets_hash = {charsets_hash}  # DO NOT CHANGE THIS LINE
    dict_hash     = {dict_hash}  # DO NOT CHANGE THIS LINE
    secrets_hash  = {secrets_hash}  # DO NOT CHANGE THIS LINE
''')


# Initial accounts file {{{1
ACCOUNTS_FILE_INITIAL_CONTENTS = dedent('''\
    # Avendesora Accounts
    # vim: filetype=python sw=4 sts=4 et ai ff=unix fileencoding={encoding} foldmethod=marker :

    # Description {section}
    # Subclass Account to create new accounts. Attributes of the subclass are
    # the account fields. Fields can be strings, lists, dictionaries or
    # Avendesora classes.
    #
    # Example:
    #     class Mortgage(Account):
    #         username = 'herbie962'
    #         account = '0036727963'
    #         passcode = PasswordRecipe('12 2u 2d 2s')
    #         phone = '800-529-2345'
    #         discovery = [
    #             RecognizeURL(
    #                 'https://www.loanservicing.com/login.jsp',
    #                 script='{{account}}{{tab}}{{passcode}}{{return}}',
    #             ),
    #         ]

    # Imports {section}
    from avendesora import (
        # Basics
        Account, StealthAccount, Hide, Hidden, GPG, Script,

        # Character sets
        exclude, LOWERCASE, UPPERCASE, LETTERS, DIGITS, ALPHANUMERIC,
        HEXDIGITS, PUNCTUATION, SYMBOLS, WHITESPACE, PRINTABLE, DISTINGUISHABLE,

        # Secrets
        Password, Passphrase, PIN, Question, MixedPassword, PasswordRecipe,
        BirthDate, OTP,

        # Account Discovery
        RecognizeAll, RecognizeAny, RecognizeCWD, RecognizeEnvVar,
        RecognizeFile, RecognizeHost, RecognizeNetwork, RecognizeTitle,
        RecognizeURL, RecognizeUser,
    )
    try:
        import Scrypt
    except ImportError:
        # You need to install scrypt using 'pip install scrypt' to use Scrypt
        pass

    # Shared Values {section}
    # These values are shared across all accounts defined in this file.
    #
    # Never disclose the master seed to anyone you are not collaborating with.
    # Never give it to your collaborator using an insecure channel, such as
    # email.  If you have encrypted this file with your colloborator's public
    # GPG key, is best to simply send the encrypted file to your
    # collaborator the first time. After than, you can send the account
    # informaiton in clear text as long as you are careful not to expose the
    # master_seed.
    master_seed = Hidden({master_seed}, secure=True)
    gpg_ids = {gpg_ids}


    # Accounts
''')

# Initial stealth_accounts file {{{1
STEALTH_ACCOUNTS_FILE_INITIAL_CONTENTS = dedent('''\
    # Avendesora Stealth Accounts
    # vim: filetype=python sw=4 sts=4 et ai ff=unix fileencoding={encoding} foldmethod=marker :

    # Description {section}
    # Stealth accounts are accounts where the class name is not taken to be the
    # account name, instead the user is interactively asked for the account
    # name. In this way there is no record that the account actually exists.

    # Imports {section}
    from avendesora import (
        # Account
        StealthAccount,

        # Character sets
        exclude, LOWERCASE, UPPERCASE, LETTERS, DIGITS, ALPHANUMERIC,
        HEXDIGITS, PUNCTUATION, WHITESPACE, PRINTABLE, DISTINGUISHABLE,

        # Secrets
        Password, PasswordRecipe, Passphrase, PIN, Hidden
    )

    # Accounts
    # Alphanumeric passwords {section}
    class Anum4(StealthAccount):
        passcode = Password(length=4, alphabet=DISTINGUISHABLE)

    class Anum6(StealthAccount):
        passcode = Password(length=6, alphabet=DISTINGUISHABLE)

    class Anum8(StealthAccount):
        passcode = Password(length=8, alphabet=DISTINGUISHABLE)

    class Anum10(StealthAccount):
        passcode = Password(length=10, alphabet=DISTINGUISHABLE)

    class Anum12(StealthAccount):
        aliases = 'anum'
        passcode = Password(length=12, alphabet=DISTINGUISHABLE)

    class Anum14(StealthAccount):
        passcode = Password(length=14, alphabet=DISTINGUISHABLE)

    class Anum16(StealthAccount):
        passcode = Password(length=16, alphabet=DISTINGUISHABLE)

    class Anum18(StealthAccount):
        passcode = Password(length=18, alphabet=DISTINGUISHABLE)

    class Anum20(StealthAccount):
        passcode = Password(length=20, alphabet=DISTINGUISHABLE)

    class Anum24(StealthAccount):
        passcode = Password(length=24, alphabet=DISTINGUISHABLE)

    class Anum32(StealthAccount):
        passcode = Password(length=32, alphabet=DISTINGUISHABLE)

    class Anum48(StealthAccount):
        passcode = Password(length=48, alphabet=DISTINGUISHABLE)

    class Anum64(StealthAccount):
        passcode = Password(length=64, alphabet=DISTINGUISHABLE)


    # Ascii passwords {section}
    class Char4(StealthAccount):
        passcode = Password(length=4, alphabet=ALPHANUMERIC+PUNCTUATION)

    class Char6(StealthAccount):
        passcode = Password(length=6, alphabet=ALPHANUMERIC+PUNCTUATION)

    class Char8(StealthAccount):
        passcode = Password(length=8, alphabet=ALPHANUMERIC+PUNCTUATION)

    class Char10(StealthAccount):
        passcode = Password(length=10, alphabet=ALPHANUMERIC+PUNCTUATION)

    class Char12(StealthAccount):
        aliases = 'char'
        passcode = Password(length=12, alphabet=ALPHANUMERIC+PUNCTUATION)

    class Char14(StealthAccount):
        passcode = Password(length=14, alphabet=ALPHANUMERIC+PUNCTUATION)

    class Char16(StealthAccount):
        passcode = Password(length=16, alphabet=ALPHANUMERIC+PUNCTUATION)

    class Char18(StealthAccount):
        passcode = Password(length=18, alphabet=ALPHANUMERIC+PUNCTUATION)

    class Char20(StealthAccount):
        passcode = Password(length=20, alphabet=ALPHANUMERIC+PUNCTUATION)

    class Char24(StealthAccount):
        passcode = Password(length=24, alphabet=ALPHANUMERIC+PUNCTUATION)

    class Char32(StealthAccount):
        passcode = Password(length=32, alphabet=ALPHANUMERIC+PUNCTUATION)

    class Char48(StealthAccount):
        passcode = Password(length=48, alphabet=ALPHANUMERIC+PUNCTUATION)

    class Char64(StealthAccount):
        alias = 'extreme'
        passcode = Password(length=64, alphabet=ALPHANUMERIC+PUNCTUATION)


    # Hex passwords {section}
    class Hex4(StealthAccount):
        passcode = Password(length=4, alphabet=HEXDIGITS, prefix='0x')

    class Hex8(StealthAccount):
        passcode = Password(length=8, alphabet=HEXDIGITS, prefix='0x')

    class Hex16(StealthAccount):
        passcode = Password(length=16, alphabet=HEXDIGITS, prefix='0x')

    class Hex24(StealthAccount):
        passcode = Password(length=24, alphabet=HEXDIGITS, prefix='0x')

    class Hex32(StealthAccount):
        passcode = Password(length=32, alphabet=HEXDIGITS, prefix='0x')

    class Hex48(StealthAccount):
        passcode = Password(length=48, alphabet=HEXDIGITS, prefix='0x')

    class Hex64(StealthAccount):
        passcode = Password(length=64, alphabet=HEXDIGITS, prefix='0x')

    class Hex96(StealthAccount):
        passcode = Password(length=96, alphabet=HEXDIGITS, prefix='0x')


    # PINs {section}
    class Pin2(StealthAccount):
        passcode = PIN(length=2)

    class Pin3(StealthAccount):
        passcode = PIN(length=3)

    class Pin4(StealthAccount):
        aliases = 'pin'
        passcode = PIN(length=4)

    class Pin5(StealthAccount):
        passcode = PIN(length=5)

    class Pin6(StealthAccount):
        passcode = PIN(length=6)

    class Pin7(StealthAccount):
        passcode = PIN(length=7)

    class Pin8(StealthAccount):
        aliases = 'num'
        passcode = PIN(length=8)

    class Pin10(StealthAccount):
        passcode = PIN(length=10)

    class Pin12(StealthAccount):
        passcode = PIN(length=12)

    class Pin14(StealthAccount):
        passcode = PIN(length=14)

    class Pin16(StealthAccount):
        passcode = PIN(length=16)

    class Pin18(StealthAccount):
        passcode = PIN(length=18)

    class Pin20(StealthAccount):
        passcode = PIN(length=20)

    class Pin24(StealthAccount):
        passcode = PIN(length=24)

    class Pin32(StealthAccount):
        passcode = PIN(length=32)

    class Pin48(StealthAccount):
        passcode = PIN(length=48)

    class Pin64(StealthAccount):
        passcode = PIN(length=64)


    # Passphrases {section}
    class Word1(StealthAccount):
        aliases = 'word'
        passcode = Passphrase(length=1)

    class Word2(StealthAccount):
        aliases = 'pair'
        passcode = Passphrase(length=2)

    class Word3(StealthAccount):
        passcode = Passphrase(length=3)

    class Word4(StealthAccount):
        aliases = 'words xkcd'
        passcode = Passphrase(length=4)

    class Word5(StealthAccount):
        passcode = Passphrase(length=5)

    class Word6(StealthAccount):
        passcode = Passphrase(length=6)

    class Word7(StealthAccount):
        passcode = Passphrase(length=7)

    class Word8(StealthAccount):
        passcode = Passphrase(length=8)

    class Word10(StealthAccount):
        passcode = Passphrase(length=10)

    class Word12(StealthAccount):
        passcode = Passphrase(length=12)

    class Word16(StealthAccount):
        passcode = Passphrase(length=16)

    class Word20(StealthAccount):
        passcode = Passphrase(length=20)

    class Word24(StealthAccount):
        passcode = Passphrase(length=24)

    class Word32(StealthAccount):
        passcode = Passphrase(length=32)


    # Web passwords {section}
    class Web6(StealthAccount):
        passcode = PasswordRecipe('6 u d s')

    class Web8(StealthAccount):
        passcode = PasswordRecipe('8 u d s')

    class Web10(StealthAccount):
        passcode = PasswordRecipe('10 2u 2d 2s')

    class Web12(StealthAccount):
        aliases = 'web'
        passcode = PasswordRecipe('12 2u 2d 2s')

    class Web14(StealthAccount):
        passcode = PasswordRecipe('14 2u 2d 2s')

    class Web16(StealthAccount):
        passcode = PasswordRecipe('16 3u 3d 3s')

    class Web18(StealthAccount):
        passcode = PasswordRecipe('18 3u 3d 3s')

    class Web20(StealthAccount):
        passcode = PasswordRecipe('20 3u 3d 3s')

    class Web24(StealthAccount):
        passcode = PasswordRecipe('24 4u 4d 4s')

    class Web32(StealthAccount):
        passcode = PasswordRecipe('32 5u 5d 5s')

    class Web48(StealthAccount):
        passcode = PasswordRecipe('48 6u 6d 6s')

    class Web64(StealthAccount):
        passcode = PasswordRecipe('64 8u 8d 8s')
''')

# Initial user key file {{{1
USER_KEY_FILE_INITIAL_CONTENTS = dedent('''\
    # Avendesora Personal Key
    # vim: filetype=python sw=4 sts=4 et ai ff=unix fileencoding={encoding} :
    #
    # This is your personal encryption key.
    # DO NOT CHANGE THIS KEY.  DO NOT SHARE THIS KEY.
    user_key = ({user_key})
''')

# Archive file header {{{1
ARCHIVE_FILE_CONTENTS = dedent('''\
    # Avendesora Archive File
    # vim: filetype=python sw=4 sts=4 et ai ff=unix fileencoding={encoding} :
    #
    # This file allows you to recover your secrets if Avendesora is not
    # available.  Except for the master seeds it contains all of the
    # information for all of the accounts. The secrets have been generated and
    # hidden using base64 encoding. To decode a hidden argument of the form:
    # Hidden('<hidden_arg>'), you can use:
    #
    #     > avendesora reveal
    #     hidden text: <hidden_arg>
    #
    # or
    #
    #     > base64 -d - < <filename>
    #
    # where <filename> contains <hidden_arg>.
    #

    from avendesora import (
        Hidden, Question, Script, OTP,
        RecognizeAll, RecognizeAny, RecognizeTitle, RecognizeURL, RecognizeCWD,
        RecognizeHost, RecognizeUser, RecognizeEnvVar, RecognizeNetwork,
        RecognizeFile
    )

    CREATED = '{date}'
    ACCOUNTS = {{
    {accounts}
    }}
''')

# Account list file {{{1
ACCOUNT_LIST_FILE_CONTENTS = dedent('''\
    # The list of files that contain accounts. The order of the files is
    # immaterial, except that first file given is the default file, meaning that
    # the add account will add a new account to the first files specified unless
    # a different files is specified explicitly.

    accounts_files = {accounts_files}
''')


# Fields {{{1
# Fields reserved for use by Avendesora
# They are not displayed by the values command.
TOOL_FIELDS = '''
    NAME aliases default discovery browser default_url
    master_seed account_seed
'''.split()



# Error

# License {{{1
# Copyright (C) 2016-18 Kenneth S. Kundert
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses/.

# Imports {{{1
from inform import Error


# Exceptions {{{1
class PasswordError(Error):
    """Password error.

    This exception subclasses
    `Inform.Error <https://github.com/KenKundert/inform#exception>`_.
    """
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
